/*
	X Windows test program
	Neuron Computing Inc.
*/

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <stdlib.h>
#include <stdio.h>
#include <sys/unistd.h>

/* Font names */
#define	FAILFONT "fixed"
#ifndef BIFONT
	#define BIGFONT "Sans Serif"
#endif
#ifndef SMALLFONT
	#define SMALLFONT BIGFONT
#endif

/* Global resources */
XFontStruct* g_fontBig;
XFontStruct* g_fontSmall;
unsigned long g_colorBlack;
unsigned long g_colorWhite;

/* Error messages */
char* g_formatErrorLoadFont = "%s: can't load font `%s', using `%s'\n";
char* g_formatErrorLoadDefaultFont = "%s: can't load default font `%s', giving up\n";

/* Global application variables */
static char* s_display_server = "127.0.0.1:0.0";

char* g_appName;
Display* g_d;
int g_s;
Window g_w;

/* Resources */
void getColors()
{
	XColor c;
	XColor e;
	Status s;

	s = XAllocNamedColor(g_d, DefaultColormap(g_d, g_s), "Black", &c, &e);
	if (s != (Status)1)
	{
		fprintf(stderr, "%s: warning: can't allocate color `Black'\n");
		g_colorBlack = BlackPixel(g_d, g_s);
	}
	else
	{
		g_colorBlack = c.pixel;
	}
	s = XAllocNamedColor(g_d, DefaultColormap(g_d, g_s), "White", &c, &e);
	if (s != (Status)1)
	{
		fprintf(stderr, "%s: can't allocate color `White'\n");
		g_colorWhite = WhitePixel(g_d, g_s);
	}
	else
	{
		g_colorWhite = c.pixel;
	}
}

void getFonts()
{
	g_fontBig = XLoadQueryFont(g_d, BIGFONT);
	if (g_fontBig == NULL)
	{
		fprintf(stderr, g_formatErrorLoadFont, g_appName, BIGFONT, FAILFONT);
		g_fontBig = XLoadQueryFont(g_d, FAILFONT);
		if (g_fontBig == (XFontStruct *)NULL)
		{
			fprintf(stderr, g_formatErrorLoadDefaultFont, g_appName, FAILFONT);
			exit(1);
		}
	}
	g_fontSmall = XLoadQueryFont(g_d, SMALLFONT);
	if (g_fontSmall == NULL)
	{
		fprintf(stderr, g_formatErrorLoadFont, g_appName, SMALLFONT, FAILFONT);
		g_fontSmall = XLoadQueryFont(g_d, FAILFONT);
		if (g_fontSmall == NULL)
		{
			fprintf(stderr, g_formatErrorLoadDefaultFont, g_appName, FAILFONT);
			exit(1);
		}
	}
}

GC getGC(Window w)
{
	XGCValues gcv;
	gcv.foreground = g_colorBlack;
	gcv.background = g_colorWhite;
	return XCreateGC(g_d, w, GCForeground | GCBackground, &gcv);
}

/* Create main frame window */
void createMainFrame()
{
	XSetWindowAttributes xswa;

	xswa.background_pixel = g_colorWhite;
	xswa.border_pixel = g_colorBlack;
	xswa.backing_store = WhenMapped;
	int mask = CWBackPixel | CWBorderPixel | CWBackingStore;

	g_w = XCreateWindow(g_d, RootWindow(g_d, g_s),
					10, 10, 320, 240,
					3, CopyFromParent,
					InputOutput, CopyFromParent, mask, &xswa);
}

/* Event handlers */
void paint_grid(Window w, GC gc, int max, int step)
{
	int i;
	for (i=0; i<=max; i+=step)
	{
		XDrawLine(g_d, w, gc, i, 0, i, max);
		XDrawLine(g_d, w, gc, 0, i, max, i);
	}
}

void comp_paint(XExposeEvent* e)
{
	if (e->window == g_w)
	{
		GC gc = getGC(e->window);
		paint_grid(e->window, gc, 500, 20);
		XFreeGC(g_d, gc);
	}
}

void comp_mouse_down(XButtonEvent* e)
{
}

/* Event logging */
char* szEvents[] = {
	"0",
	"1",
	"KeyPress",
	"KeyRelease",
	"ButtonPress",
	"ButtonRelease",
	"MotionNotify",
	"EnterNotify",
	"LeaveNotify",
	"FocusIn",
	"FocusOut",
	"KeymapNotify",
	"Expose",
	"GraphicsExpose",
	"NoExpose",
	"VisibilityNotify",
	"CreateNotify",
	"DestroyNotify",
	"UnmapNotify",
	"MapNotify",
	"MapRequest",
	"ReparentNotify",
	"ConfigureNotify",
	"ConfigureRequest",
	"GravityNotify",
	"ResizeRequest",
	"CirculateNotify",
	"CirculateRequest",
	"PropertyNotify",
	"SelectionClear",
	"SelectionRequest",
	"SelectionNotify",
	"ColormapNotify",
	"ClientMessage",
	"MappingNotify",
	"Unknown",
};

void processEvent(XEvent* e)
{
	printf("Event %s\n", szEvents[e->type]);
	/* ev.xexpose.window, ev.xbutton.x, ev.xbutton.y */
	switch (e->type)
	{
	/* Keyboard */
	case KeyPress:
	case KeyRelease:
		break;
	/* Mouse */
	case ButtonPress:
		comp_mouse_down((XButtonEvent*)e);
		break;
	case ButtonRelease:
	case MotionNotify:
	case EnterNotify:
	case LeaveNotify:
	/* Focus */
	case FocusIn:
	case FocusOut:
	case KeymapNotify:
		break;
	case Expose:
		if (e->xexpose.count == 0)
			comp_paint((XExposeEvent*)e);
		break;
	case GraphicsExpose:
	case NoExpose:
	case VisibilityNotify:
	case CreateNotify:
		break;
	case DestroyNotify:
		exit(1);
		break;
	case UnmapNotify:
	case MapNotify:
	case MapRequest:
	case ReparentNotify:
	case ConfigureNotify:
	case ConfigureRequest:
	case GravityNotify:
	case ResizeRequest:
	case CirculateNotify:
	case CirculateRequest:
	case PropertyNotify:
	case SelectionClear:
	case SelectionRequest:
	case SelectionNotify:
	case ColormapNotify:
	case ClientMessage:
	case MappingNotify:
	default: break;
	}
}

/* Event loop */
Bool evpred(Display* d, XEvent* e, char* a)
{
	return True;
}

void doTimeout()
{
}

void eventLoop()
{
	XEvent e;
	for (;;)
	{
		if (XCheckIfEvent(g_d, &e, evpred, NULL))
		{
			processEvent(&e);
		}
		else
		{
			sleep(1);
			doTimeout();
		}
	}
}

/* Clean up */
void shutDown()
{
	XFreeFont(g_d, g_fontBig);
	XFreeFont(g_d, g_fontSmall);
	XDestroyWindow(g_d, g_w);
	XCloseDisplay(g_d);
}

/* main */
int main(int argc, char ** argv)
{
	/* Initialize application */
	g_appName = argv[0];

	g_d = XOpenDisplay(s_display_server);
	if (g_d == NULL)
	{
		fprintf(stderr, "%s: can't open display `%s'\n", s_display_server);
		exit(1);
	}
	g_s = DefaultScreen(g_d);

	/* Load resources */
	getColors();
	getFonts();
	
	/* Create main frame */
	createMainFrame();

	unsigned long MOUSE_MASK = ButtonPressMask | ButtonReleaseMask |
								EnterWindowMask | LeaveWindowMask;
	unsigned long KEY_MASK = KeyPressMask | KeyReleaseMask;
	unsigned long MASK = ExposureMask | MOUSE_MASK | KEY_MASK |
						StructureNotifyMask | SubstructureNotifyMask |
						VisibilityChangeMask | FocusChangeMask |
						PropertyChangeMask | ColormapChangeMask;
	Window root = RootWindow(g_d, g_s);
	XSelectInput(g_d, g_w, MASK);
	XMapWindow(g_d, g_w);

	/* Process events */
	eventLoop();

	/* Clean up and exit */
	shutDown();
	exit(0);

	return 0;
}
